from .device_name import DevName, expand_name_variants
from .mask import create_mask, check_mask
from .device_group import DevGroup, DevSubgroup
from .generator import DeviceDataGenerator
from .reader import DeviceDataReader



import re
from dataclasses import dataclass

from ..._logger import logger

_peripheral_name_pattern = re.compile(
    r"(?P<name>.*?)"
    r"(?P<instance>\d{1,2})?"
)


class WrongPeripheralNamePattern(Exception):
    def __init__(self, name_str: str):
        super().__init__(
            f"Peripheral name '{name_str}' does not match expected pattern"
        )


@dataclass(slots=True, frozen=True)
class PeripheralName:
    name: str
    instance: int

    @classmethod
    def from_str(cls, s: str):
        try:
            match = _peripheral_name_pattern.fullmatch(s.lower())
            if match is None:
                raise WrongPeripheralNamePattern(s)
            instance = int(match.group("instance")) if match.group("instance") else 0
            name = match.group("name")
            return cls(name, instance)

        except WrongPeripheralNamePattern:
            logger.exception("Can't parse peripheral name because of pattern mismatch")
            raise



from dataclasses import dataclass
from typing import Sequence

from .device_name import DevName


@dataclass
class DevSubgroup:
    line: str
    pin: Sequence[str]
    memory: Sequence[str]


@dataclass
class DevGroup:
    name: str
    family: str
    subgroups: list[DevSubgroup]

    def includes_device(self, device: DevName) -> bool:
        if device.family != self.family:
            return False

        for subgroup in self.subgroups:
            if (
                device.line == subgroup.line
                and device.pin in subgroup.pin
                and device.memory in subgroup.memory
            ):
                return True
        return False



from ._db import stm32_memory_db
from ..device_name import DevName
from ._models import MemSeg
from ..mask import check_mask


def get_segments_from_db(dev: DevName) -> list[MemSeg]:
    found_segments: list[MemSeg] = []

    for mask in stm32_memory_db.keys():
        if check_mask(mask, dev):
            found_segments += stm32_memory_db[mask]

    return found_segments



import copy

from ._utils import get_segments_from_db
from ._models import MemSegType, MemSegMasked, MemSeg, MemMap
from ..device_name import DevName
from ..mask import create_mask
from ..cubemx import DevData


class MemoryDataManager:
    def __init__(self, devices_data: list[DevData]):
        self._devices_data = devices_data

    def __get_devices_by_segment(self) -> dict[MemSeg, list[DevName]]:
        devices_by_segment = dict()

        for dev_data in self._devices_data:
            for dev_memory, dev in zip(dev_data.memory_variants, dev_data.devices):
                dev_segments = get_segments_from_db(dev)
                for dev_seg in dev_segments:
                    if dev_seg.seg_type == MemSegType.FLASH:
                        segment = MemSeg(
                            MemSegType.FLASH, dev_seg.start, dev_memory.flash_size
                        )
                    else:
                        segment = copy.deepcopy(dev_seg)

                    if segment in devices_by_segment.keys():
                        devices_by_segment[segment].add(dev)
                    else:
                        devices_by_segment[segment] = {dev}
        return devices_by_segment

    def get_memory_map(self) -> MemMap:
        devices_by_segment = self.__get_devices_by_segment()
        masked_segments: list[MemSegMasked] = []

        for segment in devices_by_segment.keys():
            mask = create_mask(devices=devices_by_segment[segment])
            masked_segments.append(
                MemSegMasked(segment.seg_type, segment.start, segment.size, mask)
            )

        masked_segments.sort(key=lambda s: (s.seg_type.value, s.size))

        return MemMap(segments=masked_segments)



import os
from pathlib import Path
import json
from dataclasses import asdict

from .device_group import DevGroup
from .cubemx import FamiliesFile, ValidDevices, DevData, MCUFile
from .device_data import STM32Data

from .memory import MemoryDataManager, MemMap
from .peripherals import PeripheralDataManager, PeripheralMasked


class DeviceDataGenerator:
    def __init__(self, cubemx_db_folder: str | Path, output_folder: str | Path):
        self._db_folder = os.path.abspath(cubemx_db_folder)
        self._mcu_folder = os.path.join(self._db_folder, "mcu")
        self._ip_folder = os.path.join(self._mcu_folder, "IP")
        self._output_folder = os.path.abspath(output_folder)

        self._valid_devices: ValidDevices | None = None
        self._memory_map: MemMap | None = None
        self._peripherals: list[PeripheralMasked] | None = None

        self._data: STM32Data | None = None

    def generate(self, dev_group: DevGroup):
        families_file = FamiliesFile(os.path.join(self._mcu_folder, "families.xml"))
        self._valid_devices = families_file.extract_valid_devices(dev_group)

        devices_data: list[DevData] = []
        for cubemx_name in self._valid_devices.cubemx_names:
            filepath = os.path.join(self._mcu_folder, f"{cubemx_name}.xml")
            mcu_file = MCUFile(filepath)
            devices_data.append(mcu_file.extract_all_data())

        memory_data_collector = MemoryDataManager(devices_data=devices_data)
        self._memory_map = memory_data_collector.get_memory_map()

        peripheral_data_collector = PeripheralDataManager(devices_data=devices_data)
        self._peripherals = peripheral_data_collector.get_all_peripherals()

        self.__compose_data()
        output_filepath = os.path.join(self._output_folder, f"{dev_group.name}.json")
        self.__save_data_as_json(output_filepath)

    def __compose_data(self):
        devices = [dev.model for dev in self._valid_devices.devices]

        self._data = STM32Data(
            devices=devices,
            memory=self._memory_map,
            peripherals=self._peripherals,
        )

    def __save_data_as_json(self, filepath: str):
        data_dict = asdict(self._data)
        data_json = json.dumps(data_dict, indent=4)
        with open(filepath, "w") as f:
            f.write(data_json)



import os
from pathlib import Path
import json

from .device_name import DevName
from .mask import check_mask
from .._logger import logger

from .memory import MemSeg
from .peripherals import Peripheral


class FileNotFoundForDevice(Exception):
    def __init__(self, device: str):
        super().__init__(f"No file with data for '{device}' device")


class DeviceDataReader:
    def __init__(self, folder: str | Path):
        self._data_folder = folder
        self._device: DevName | None = None

        self._data: dict | None = None

    def set_device(self, dev_name: str):
        self._device = DevName.from_str(dev_name)
        self.__find_data_file()

    def __find_data_file(self):
        try:
            filenames = [
                f for f in os.listdir(self._data_folder) if f.find(".json") > 0
            ]
            for filename in filenames:
                filepath = os.path.join(self._data_folder, filename)
                with open(filepath, "r") as f:
                    data = json.load(f)
                if self._device.model in data["devices"]:
                    self._data = data
                    return
            raise FileNotFoundForDevice(self._device.model)

        except FileNotFoundForDevice:
            logger.exception("Can't find device data file for provided device")
            raise

    def get_memory_segments(self) -> list[MemSeg]:
        return [
            MemSeg.from_dict(seg_data)
            for seg_data in self._data["memory"]["segments"]
            if check_mask(seg_data["mask"], self._device)
        ]

    def get_peripheral(self, name: str, instance: int) -> Peripheral | None:
        for periph_data in self._data["peripherals"]:
            if name == periph_data["name"] and instance in periph_data["instances"]:
                if check_mask(periph_data["mask"], self._device):
                    return Peripheral.from_dict(periph_data)
        return None



import copy

from ..device_name import DevName
from ..mask import create_mask
from ..cubemx import DevData, DevPeripheral
from ._utils import get_peripheral_from_db
from ._models import Peripheral, PeripheralDef, PeripheralMasked


_stm32_ignored_peripherals = [
    "fatfs",
    "freertos",
    "libjpeg",
    "mbedtls",
    "pdm2pcm",
    "usb_device",
    "usb_host",
    "usb_otg_fs",
    "usb_otg_hs",
    "wwdg",
    "nvic",
]


class PeripheralDataManager:
    def __init__(self, devices_data: list[DevData]):
        self._devices_data = devices_data

        self._peripherals_by_device: dict[str, list[Peripheral]] = dict()

    def get_all_peripherals(self) -> list[PeripheralMasked]:
        self.__do_initial_processing()

        merged_by_version = self.__merge_by_version()
        masked_peripherals = self.__create_masked_peripherals(merged_by_version)

        final_peripherals = self.__merge_by_devices(masked_peripherals)

        final_peripherals.sort(key=lambda s: (s.name, s.instances))
        return final_peripherals

    def __merge_by_devices(
        self,
        peripherals: list[PeripheralMasked],
    ) -> list[PeripheralMasked]:
        for i in range(len(peripherals) - 1, -1, -1):
            for j in range(i - 1, -1, -1):
                p_i, p_j = peripherals[i], peripherals[j]
                if self.__can_merge_by_devices(p_i, p_j):
                    peripherals[j] = PeripheralMasked(
                        name=p_j.name,
                        instances=p_j.instances + p_i.instances,
                        version=p_j.version,
                        tags=p_i.tags,
                        mask=p_i.mask,
                    )
                    peripherals.pop(i)
                    break

        return peripherals

    @staticmethod
    def __can_merge_by_devices(p1: PeripheralMasked, p2: PeripheralMasked) -> bool:
        return (
            p1.name == p2.name
            and p1.version == p2.version
            and p1.tags == p2.tags
            and p1.mask == p2.mask
        )

    @staticmethod
    def __create_masked_peripherals(
        merged_by_version: dict[Peripheral, set[str]],
    ) -> list[PeripheralMasked]:
        masked_peripherals: list[PeripheralMasked] = []
        for p in merged_by_version.keys():
            mask = create_mask(cubemx_devices=list(merged_by_version[p]))
            masked_peripherals.append(
                PeripheralMasked(
                    name=p.name,
                    instances=p.instances,
                    version=p.version,
                    tags=p.tags,
                    mask=mask,
                )
            )
        return masked_peripherals

    def __merge_by_version(self) -> dict[Peripheral, set[str]]:
        merged_by_version: dict[Peripheral, set[str]] = dict()
        for dev in self._peripherals_by_device.keys():
            for p in self._peripherals_by_device[dev]:
                if p not in merged_by_version.keys():
                    merged_by_version[p] = {dev}
                else:
                    merged_by_version[p].add(dev)
        return merged_by_version

    @staticmethod
    def __remove_ignored_peripherals(
        input_peripherals: list[DevPeripheral],
    ) -> list[DevPeripheral]:
        output_peripherals = []
        for p in input_peripherals:
            if p.name not in _stm32_ignored_peripherals:
                output_peripherals.append(p)
        return output_peripherals

    def __do_initial_processing(self):
        for dev_data in self._devices_data:
            peripherals = self.__remove_ignored_peripherals(dev_data.peripherals)
            peripherals_embx = self.__apply_embx_versions(
                peripherals, dev_data.devices[0]
            )
            self._peripherals_by_device[dev_data.cubemx_device] = peripherals_embx

    @staticmethod
    def __apply_embx_versions(
        cube_peripherals: list[DevPeripheral], dev: DevName
    ) -> list[Peripheral]:
        output = []
        for p in cube_peripherals:
            db_elem: PeripheralDef = get_peripheral_from_db(p.name, p.instance, dev)

            output.append(
                Peripheral(p.name.lower(), (p.instance,), db_elem.version, db_elem.tags)
            )
        return output



from ..._logger import logger
from ..device_name import DevName
from ...utils.base_xml_parser import BaseXMLParser
from ..device_group import DevGroup
from ._models import ValidDevices


class FamiliesFile(BaseXMLParser):
    def __init__(self, filepath: str):
        super().__init__(filepath=filepath)

    def extract_valid_devices(self, dev_group: DevGroup) -> ValidDevices:
        mcu_elems = self._root.findall("Family/SubFamily/Mcu")

        ref_names = []
        cubemx_names = set()

        for mcu_elem in mcu_elems:
            ref_name = mcu_elem.get("RefName").lower()
            cubemx_name = mcu_elem.get("Name")
            parsed_name = DevName.from_str(ref_name)
            if dev_group.includes_device(parsed_name):
                ref_names.append(ref_name)
                cubemx_names.add(cubemx_name)

        parsed_names = [DevName.from_str(dev) for dev in sorted(ref_names)]

        return ValidDevices(
            cubemx_names=sorted(cubemx_names),
            devices=parsed_names,
        )



from xml.etree import ElementTree
from xml.etree.ElementTree import Element
import re

from .._logger import logger


class NamespaceNameAlreadyUsed(Exception):
    def __init__(self, tag: str, name: str, uri_new: str, uri_old: str):
        super().__init__(
            f"Namespace '{name}' already mapped to '{uri_old}', "
            f"can't remap to '{uri_new}' from element '{tag}'"
        )


class BaseXMLParser:
    def __init__(self, filepath: str):
        self._tree = ElementTree.parse(filepath)
        self._root = self._tree.getroot()
        self._ns: dict = {}

    def register_namespace(self, elem: Element, name: str):
        try:
            uri = self._get_namespace_uri(elem)

            if uri is None:
                logger.warning(f"Namespace not found in element '{elem.tag}'")
                return

            if name not in self._ns.keys():
                self._ns[name] = uri
            elif self._ns[name] != uri:
                raise NamespaceNameAlreadyUsed(elem.tag, name, uri, self._ns[name])

        except NamespaceNameAlreadyUsed:
            logger.exception(
                "Namespace with same name but different uri is already registered"
            )
            raise

    @staticmethod
    def _get_namespace_uri(elem: Element) -> str | None:
        m = re.match(r"\{(.*)}", elem.tag)
        return m.group(1) if m else None



from ._models import PeripheralDef
from ..device_name import DevName
from ..mask import check_mask
from ._db import stm32_peripheral_db
from ..._logger import logger


class NoPeripheralVersionInDB(Exception):
    def __init__(self, peripheral_name: str, instance: int, dev: DevName):
        super().__init__(
            f"Peripheral version not found in database: "
            f"name='{peripheral_name}', instance={instance}, device='{dev.model}'"
        )


def get_peripheral_from_db(
    peripheral_name: str, instance: int, dev: DevName
) -> PeripheralDef | None:
    if peripheral_name in stm32_peripheral_db.keys():
        periph_db = stm32_peripheral_db[peripheral_name]
        for mask in periph_db.keys():
            if check_mask(mask, dev):
                for periph_case in periph_db[mask]:
                    if instance in periph_case.instances:
                        return periph_case

    try:
        raise NoPeripheralVersionInDB(peripheral_name, instance, dev)
    except NoPeripheralVersionInDB:
        logger.exception("Can't find peripheral version in database")
        raise



from dataclasses import dataclass
import re
from itertools import product

from .._logger import logger


class WrongSTM32NamePattern(Exception):
    def __init__(self, name_str: str):
        super().__init__(f"STM32 name '{name_str}' does not match expected pattern")


_stm32_name_pattern = re.compile(
    r"stm32"
    r"(?P<family>f0|f1|f2|f3|f4|f7|l0|l1|l4|l4\+|l5|u0|u3|u5|n6|h5|h7|g0|g4|wb|wb0|wba|wl|c0|mp1|mp2|wl3)"
    r"(?:(?P<line>[0-9a-z]{2})"
    r"(?:(?P<pin>[dyfegkthscurjmovqzaibnxplw])"
    r"(?:(?P<memory>[0123456789abzcdefghiyj])"
    r"(?:(?P<package>[bdghijkmpqtuvyxflcaes])"
    r"(?:(?P<temperature>[673abcdx])"
    r"(?P<variant>.*)?)?)?)?)?)?"
)


@dataclass(slots=True, frozen=True)
class DevName:
    family: str
    line: str
    pin: str
    memory: str
    package: str
    temperature: str
    variant: str

    @property
    def model(self) -> str:
        return f"stm32{self.family}{self.line}{self.pin}{self.memory}{self.package}"

    @classmethod
    def from_str(cls, s: str):
        try:
            match = _stm32_name_pattern.fullmatch(s.lower())
            if match is None:
                raise WrongSTM32NamePattern(s)

            def get_field(field_name: str, remove_x: bool = False) -> str:
                d = match.group(field_name) or ""
                if remove_x:
                    d = "" if d == "x" else d
                return d

            return cls(
                family=get_field("family"),
                line=get_field("line"),
                pin=get_field("pin"),
                memory=get_field("memory"),
                package=get_field("package", remove_x=True),
                temperature=get_field("temperature", remove_x=True),
                variant=get_field("variant"),
            )

        except WrongSTM32NamePattern:
            logger.exception("Can't parse STM32 name because of pattern mismatch")
            raise


# STM32G473P(B-C-E)Ix
def expand_name_variants(s: str) -> list[str]:
    parts = re.split(r"(\([^)]+\))", s)
    option_lists = []
    for part in parts:
        if part.startswith("(") and part.endswith(")"):
            # bracket part
            options = part[1:-1].split("-")
            option_lists.append(options)
        else:
            # non-bracket part
            option_lists.append([part])

    variants = ["".join(combo) for combo in product(*option_lists)]
    return variants



import copy
from xml.etree.ElementTree import Element

from ...utils.base_xml_parser import BaseXMLParser
from ..device_name import DevName, expand_name_variants
from ._models import DevData, DevMemory, DevPeripheral
from ..peripherals import PeripheralName


class MCUFile(BaseXMLParser):
    def __init__(self, filepath: str):
        super().__init__(filepath=filepath)

        self.register_namespace(self._root, "mcu")

        self._cubemx_device = self._root.get("RefName")
        self._devices = [
            DevName.from_str(dev) for dev in expand_name_variants(self._cubemx_device)
        ]

    def __extract_memory(self) -> list[DevMemory]:
        flash_elems = self._root.findall("mcu:Flash", self._ns)
        flash_variants = [int(elem.text) * 1024 for elem in flash_elems]

        ram_elems = self._root.findall("mcu:Ram", self._ns)
        ram_variants = [int(elem.text) * 1024 for elem in ram_elems]

        ccm_elem = self._root.find("mcu:CCMRam", self._ns)
        ccm = int(ccm_elem.text) * 1024 if ccm_elem is not None else None

        output: list[DevMemory] = []
        for dev, dev_flash, dev_ram in zip(self._devices, flash_variants, ram_variants):
            output.append(DevMemory(dev_flash, dev_ram, ccm))
        return output

    def __extract_peripherals(self):
        peripherals: list[DevPeripheral] = []
        for peripheral_elem in self._root.findall("mcu:IP", self._ns):
            instance_str = peripheral_elem.get("InstanceName")
            version = peripheral_elem.get("Version")

            parsed_name = PeripheralName.from_str(instance_str)
            peripherals.append(
                DevPeripheral(
                    name=parsed_name.name,
                    instance=parsed_name.instance,
                    cubemx_version=version,
                )
            )
        return peripherals

    def extract_all_data(self) -> DevData:
        memory = self.__extract_memory()
        peripherals = self.__extract_peripherals()

        return DevData(
            cubemx_device=self._cubemx_device,
            devices=self._devices,
            memory_variants=memory,
            peripherals=peripherals,
        )



from dataclasses import dataclass

from ..device_name import DevName


@dataclass(slots=True, frozen=True)
class ValidDevices:
    cubemx_names: list[str]
    devices: list[DevName]


@dataclass(slots=True, frozen=True)
class DevMemory:
    flash_size: int
    ram_size: int
    ccm_size: int | None


@dataclass(slots=True, frozen=True)
class DevPeripheral:
    name: str
    instance: int
    cubemx_version: str


@dataclass(slots=True, frozen=True)
class DevData:
    cubemx_device: str
    devices: list[DevName]
    memory_variants: list[DevMemory]
    peripherals: list[DevPeripheral]



from dataclasses import dataclass
from enum import StrEnum


class MemSegType(StrEnum):
    FLASH = "flash"
    CCM = "ccm"
    SRAM1 = "sram1"
    SRAM2 = "sram2"
    SRAM3 = "sram3"
    BACKUP = "backup"


@dataclass(slots=True, frozen=True)
class MemSeg:
    seg_type: MemSegType
    start: int
    size: int | None

    @classmethod
    def from_dict(cls, d: dict):
        return cls(MemSegType(d["seg_type"]), d["start"], d["size"])


@dataclass(slots=True, frozen=True)
class MemSegMasked:
    seg_type: MemSegType
    start: int
    size: int
    mask: str


@dataclass(slots=True, frozen=True)
class MemMap:
    segments: list[MemSegMasked]



from typing import Sequence
from itertools import product
import re

from .device_name import DevName, expand_name_variants
from .._logger import logger

_mask_part_pattern = re.compile(
    r"\["
    r"(?P<family>.*?)"
    r"]\["
    r"(?P<lines>.*?)"
    r"]\["
    r"(?P<postfixes>.*?)"
    r"]"
)


class NoDevicesSpecifiedError(Exception):
    def __init__(self):
        super().__init__(
            f"Can't create mask: both 'devices' and 'cubemx_devices' are None"
        )


def create_mask(
    devices: Sequence[DevName] | None = None,
    cubemx_devices: Sequence[str] | None = None,
) -> str:
    if devices is None:
        if cubemx_devices is not None:
            devices = [
                DevName.from_str(n)
                for dev in cubemx_devices
                for n in expand_name_variants(dev)
            ]
        else:
            logger.error(
                f"Can't create mask: both 'devices' and 'cubemx_devices' are None"
            )
            raise NoDevicesSpecifiedError()

    families = set()
    lines = set()
    data = {}

    for name in devices:
        families.add(family := name.family)
        lines.add(line := name.line)
        postfix = f"{name.pin}{name.memory}{name.package}"
        if family not in data:
            data[family] = {}
        if line not in data[family]:
            data[family][line] = []
        data[family][line].append(postfix)

    for family in data.keys():
        for line in data[family].keys():
            data[family][line].sort()

    groups = []
    for family, line in product(sorted(families), sorted(lines)):
        # Check if family line combination exists
        if not (family in data and line in data[family]):
            continue
        merged = False
        for group in groups:
            if group[2] == data[family][line] and group[0] == family:
                group[1].append(line)
                merged = True
                break

        if not merged:
            groups.append([family, [line], data[family][line]])

    mask_parts = [
        f"[{group[0]}][{'|'.join(group[1])}][{'|'.join(group[2])}]" for group in groups
    ]
    return "|".join(mask_parts)


def check_mask(
    mask: str,
    device: DevName,
) -> bool:
    family = device.family
    line = device.line
    postfix = f"{device.pin}{device.memory}{device.package}"

    for match in _mask_part_pattern.finditer(mask):
        mask_families = match.group("family").split("|")
        mask_lines = match.group("lines").split("|")
        mask_postfixes = match.group("postfixes").split("|")

        if "*" in mask_families or family in mask_families:
            if "*" in mask_lines or line in mask_lines:
                if "*" in mask_postfixes or postfix in mask_postfixes:
                    return True
    return False



from dataclasses import dataclass, field


@dataclass(slots=True, frozen=True)
class PeripheralDef:
    instances: tuple
    version: str
    tags: tuple[str, ...] = field(default_factory=tuple)


@dataclass(slots=True, frozen=True)
class Peripheral:
    name: str
    instances: tuple
    version: str
    tags: tuple[str, ...]

    @classmethod
    def from_dict(cls, d: dict):
        return cls(d["name"], tuple(d["instances"]), d["version"], tuple(d["tags"]))


@dataclass(slots=True, frozen=True)
class PeripheralMasked:
    name: str
    instances: tuple
    version: str
    tags: tuple[str, ...]
    mask: str



from dataclasses import dataclass

from .memory import MemMap
from .peripherals import PeripheralMasked


@dataclass(frozen=True, slots=True)
class STM32Data:
    devices: list[str]
    memory: MemMap
    peripherals: list[PeripheralMasked]



from ..memory import MemSeg, MemSegType

stm32_memory_db = {
    "[f4][05|07|15|17][*]": [
        MemSeg(MemSegType.FLASH, 0x08000000, None),
        MemSeg(MemSegType.CCM, 0x10000000, 64 * 1024),
        MemSeg(MemSegType.SRAM1, 0x20000000, 112 * 1024),
        MemSeg(MemSegType.SRAM2, 0x2001C000, 16 * 1024),
        MemSeg(MemSegType.BACKUP, 0x40024000, 4 * 1024),
    ],
}
